use crate::{
    cert, client_management, config,
    schema::{self, schema},
    server, token, user_management,
};
use clap::{Parser, Subcommand};
use microrm::prelude::*;

#[derive(Debug, Parser)]
#[clap(author, version, about, long_about = None)]
struct RootArgs {
    #[clap(short, long, default_value_t = String::from("uauth.db"))]
    /// Database path
    db: String,

    #[clap(short, long, default_value_t = String::from("primary"))]
    /// Which realm to use, for non-server only
    realm: String,

    #[clap(subcommand)]
    command: Command,
}

#[derive(Debug, Subcommand)]
enum Command {
    /// database initialization
    Init,
    /// certificate management
    Cert(CertArgs),
    /// OAuth2 client management
    Client(ClientArgs),
    /// general configuration
    Config(ConfigArgs),
    /// permissions grouping management
    Group(GroupArgs),
    /// run the actual OIDC server
    Server(ServerArgs),
    /// manual token generation and inspection
    Token(TokenArgs),
    /// role management
    Role(RoleArgs),
    /// user management
    User(UserArgs),
}

impl RootArgs {
    async fn run(&self) {
        if let Command::Init = self.command {
            return self.init().await;
        }

        let storage = microrm::DB::new(schema::schema(), &self.db, microrm::CreateMode::MustExist);

        if let Err(e) = storage {
            println!("Error occured while loading database: {}", e);
            return;
        }
        let storage = storage.unwrap();

        match &self.command {
            Command::Init => unreachable!(),
            Command::Cert(v) => v.run(&self, storage).await,
            Command::Config(v) => v.run(&self, storage).await,
            Command::Client(v) => v.run(&self, storage).await,
            Command::Group(v) => v.run(&self, storage).await,
            Command::Server(v) => v.run(&self, storage).await,
            Command::Token(v) => v.run(&self, storage).await,
            Command::Role(v) => v.run(&self, storage).await,
            Command::User(v) => v.run(&self, storage).await,
        }
    }

    async fn init(&self) {
        // first check to see if the database is already vaguely set up
        let maybedb = microrm::DB::new(schema::schema(), &self.db, microrm::CreateMode::MustExist);

        if maybedb.is_ok() {
            println!("Database already initialized, not overwriting!");
            return;
        }

        log::info!("Initializing!");

        let db = microrm::DB::new(
            schema::schema(),
            &self.db,
            microrm::CreateMode::AllowNewDatabase,
        )
        .expect("Unable to initialize database!");

        // create primary realm
        db.query_interface()
            .add(&schema::Realm {
                shortname: "primary".to_string(),
            })
            .expect("couldn't add realm");
    }
}

#[derive(Debug, Subcommand)]
enum CertCommand {
    Inspect,
    Generate,
}

#[derive(Debug, Parser)]
struct CertArgs {
    #[clap(subcommand)]
    command: CertCommand,
}

impl CertArgs {
    async fn run(&self, root: &RootArgs, db: microrm::DB) {
        match &self.command {
            CertCommand::Inspect => {
                cert::inspect(&db, &root.realm);
            }
            CertCommand::Generate => {
                cert::generate(&db, &root.realm);
            }
        }
    }
}

#[derive(Debug, Subcommand)]
enum ClientCommand {
    Create { name: String },
    List,
    Inspect { name: String },
}

#[derive(Debug, Parser)]
struct ClientArgs {
    #[clap(subcommand)]
    command: ClientCommand,
}

impl ClientArgs {
    async fn run(&self, root: &RootArgs, db: microrm::DB) {
        match &self.command {
            ClientCommand::Create { name } => {
                client_management::create(&db, root.realm.as_str(), name);
            }
            ClientCommand::List => {}
            ClientCommand::Inspect { name } => {
                client_management::inspect(&db, name);
            }
        }
    }
}

#[derive(Debug, Subcommand)]
enum ConfigCommand {
    Dump,
    Load { toml_path: String },
    Set { key: String, value: String },
}

#[derive(Debug, Parser)]
struct ConfigArgs {
    #[clap(subcommand)]
    command: ConfigCommand,
}

impl ConfigArgs {
    async fn run(&self, root: &RootArgs, db: microrm::DB) {
        match &self.command {
            ConfigCommand::Dump => {
                let qi = db.query_interface();
                let config = config::Config::build_from(&qi, None);
                println!("config: {:?}", config);
            }
            ConfigCommand::Set { key, value } => {
                todo!()
            }
            ConfigCommand::Load { toml_path } => {
                let config = {
                    let qi = db.query_interface();
                    config::Config::build_from(&qi, Some(toml_path))
                };
                {
                    let qi = db.query_interface();
                    config.save(&qi);
                    drop(config);
                }
            }
        }
    }
}

#[derive(Debug, Subcommand)]
enum GroupCommand {
    Create { group_name: String },
    Members { group_name: String },
    AttachRole { group_name: String, role_name: String },
    DetachRole { group_name: String, role_name: String },
    AttachUser { group_name: String, username: String },
    DetachUser { group_name: String, username: String },
}

#[derive(Debug, Parser)]
struct GroupArgs {
    #[clap(subcommand)]
    command: GroupCommand,
}

impl GroupArgs {
    async fn run(&self, root: &RootArgs, db: microrm::DB) {
        let qi = db.query_interface();
        let realm_id = qi.get().by(schema::Realm::Shortname, root.realm.as_str()).one().unwrap().expect("no such realm").id();
        match &self.command {
            GroupCommand::Create { group_name } => {
                match qi.add(&schema::Group {
                    realm: realm_id,
                    shortname: group_name.clone(),
                }) {
                    Ok(_) => {
                        println!("Added group {}", group_name);
                    },
                    Err(e) => {
                        println!("Failed to add group: {:?}", e);
                    },
                }
            },
            GroupCommand::Members { group_name } => {
                todo!()
            },
            GroupCommand::AttachRole { group_name, role_name } => {
                let group = qi.get().by(schema::Group::Realm, &realm_id).by(schema::Group::Shortname, group_name).one().unwrap();
                let role = qi.get().by(schema::Role::Realm, &realm_id).by(schema::Role::Shortname, role_name).one().unwrap();

                match (group, role) {
                    (None, _) => {
                        println!("No such group!");
                    },
                    (_, None) => {
                        println!("No such role!");
                    },
                    (Some(group), Some(role)) => {
                        match qi.add(&schema::GroupRole {
                            group: group.id(),
                            role: role.id(),
                        }) {
                            Ok(_) => {
                                println!("Attached role {} to {}", role_name, group_name);
                            },
                            Err(e) => {
                                println!("Failed to attach role: {:?}", e);
                            }
                        }
                    },
                }

            },
            GroupCommand::DetachRole { group_name, role_name } => {
                
            },
            GroupCommand::AttachUser { group_name, username } => {
                let group = qi.get().by(schema::Group::Realm, &realm_id).by(schema::Group::Shortname, group_name).one().unwrap();
                let user = qi.get().by(schema::User::Realm, &realm_id).by(schema::User::Username, username).one().unwrap();

                match (group, user) {
                    (None, _) => {
                        println!("No such group!");
                    },
                    (_, None) => {
                        println!("No such user!");
                    },
                    (Some(group), Some(user)) => {
                        match qi.add(&schema::GroupMembership {
                            group: group.id(),
                            user: user.id(),
                        }) {
                            Ok(_) => {
                                println!("Attached user {} to {}", username, group_name);
                            },
                            Err(e) => {
                                println!("Failed to attach user: {:?}", e);
                            }
                        }
                    },
                }
            },
            GroupCommand::DetachUser { group_name, username } => {
                
            }
        }
    }
}

#[derive(Debug, Parser)]
struct ServerArgs {
    #[clap(short, long)]
    port: Option<u16>,
}

impl ServerArgs {
    async fn run(&self, root: &RootArgs, db: microrm::DB) {
        let config = config::Config::build_from(&db.query_interface(), None);
        server::run_server(db, config, self.port.unwrap_or(2114)).await
    }
}

#[derive(Debug, Subcommand)]
enum TokenCommand {
    GenerateAuth {
        #[clap(short, long)]
        client: String,
        #[clap(short, long)]
        username: String,
        #[clap(short, long)]
        scopes: String,
    },
    GenerateRefresh {
        #[clap(short, long)]
        client: String,
        #[clap(short, long)]
        username: String,
        #[clap(short, long)]
        scopes: String,
    },
    Inspect {
        token: String,
    },
}

#[derive(Debug, Parser)]
struct TokenArgs {
    #[clap(subcommand)]
    command: TokenCommand,
}

impl TokenArgs {
    async fn run(&self, root: &RootArgs, db: microrm::DB) {
        let config = config::Config::build_from(&db.query_interface(), None);
        match &self.command {
            TokenCommand::GenerateAuth {
                client,
                username,
                scopes,
            } => {
                let qi = db.query_interface();
                let realm_id = qi
                    .get()
                    .by(schema::Realm::Shortname, &root.realm)
                    .one()
                    .unwrap()
                    .expect("no such realm")
                    .id();
                let token = token::generate_auth_token(
                    &config,
                    &qi,
                    realm_id,
                    qi.get()
                        .by(schema::Client::Realm, &realm_id)
                        .by(schema::Client::Shortname, client.as_str())
                        .one()
                        .unwrap()
                        .expect("no such client")
                        .id(),
                    qi.get()
                        .by(schema::User::Realm, &realm_id)
                        .by(schema::User::Username, username.as_str())
                        .one()
                        .unwrap()
                        .expect("no such user")
                        .id(),
                    scopes.split_whitespace(),
                );
                match token {
                    Ok(t) => {
                        println!("token: {t}");
                    }
                    Err(e) => {
                        println!("could not generate token: {:?}", e);
                    }
                }
            }
            TokenCommand::GenerateRefresh {
                client,
                username,
                scopes,
            } => {}
            TokenCommand::Inspect { token } => {}
        }
    }
}

#[derive(Debug, Subcommand)]
enum RoleCommand {
    List,
    Create { name: String },
    Delete { name: String },
}

#[derive(Debug, Parser)]
struct RoleArgs {
    #[clap(subcommand)]
    command: RoleCommand,
}

impl RoleArgs {
    async fn run(&self, root: &RootArgs, db: microrm::DB) {
        let config = config::Config::build_from(&db.query_interface(), None);
        match &self.command {
            RoleCommand::List => {
                todo!()
            },
            RoleCommand::Create { name } => {
                let qi = db.query_interface();
                let realm = qi.get().by(schema::Realm::Shortname, &root.realm).one().unwrap().unwrap();
                let add_result = qi.add(&schema::Role {
                    realm: realm.id(),
                    shortname: name.clone()
                });

                match add_result {
                    Ok(_role_id) => {
                        println!("Added role {}", name);
                    }
                    Err(_) => {
                        println!("Failed to add role {}!", name);
                    },
                }
            },
            RoleCommand::Delete { name } => {
                let qi = db.query_interface();
                let realm = qi.get().by(schema::Realm::Shortname, &root.realm).one().unwrap().unwrap();
                qi.delete().by(schema::Role::Realm, &realm.id()).by(schema::Role::Shortname, name.as_str()).exec().unwrap();
            },
        }
    }
}

#[derive(Debug, Subcommand)]
enum UserCommand {
    List,
    Create {
        username: String
    },
    Auth {
        username: String,

        #[clap(short = 'p', long, parse(from_occurrences))]
        change_password: usize,
    },
    Inspect {
        username: String,
    }
}

#[derive(Debug, Parser)]
struct UserArgs {
    #[clap(subcommand)]
    command: UserCommand,
}

impl UserArgs {
    async fn run(&self, root: &RootArgs, db: microrm::DB) {
        match &self.command {
            UserCommand::List => user_management::list(&root.realm, db),
            UserCommand::Create { username } => {
                user_management::create(&root.realm, db, username.as_str())
            }
            UserCommand::Auth { username, change_password } => user_management::change_auth(
                &root.realm,
                db,
                username.as_str(),
                *change_password > 0,
            ),
            UserCommand::Inspect { username } => user_management::inspect(&root.realm, db, username.as_str()).expect("database error"),
        }
    }
}

pub fn invoked() {
    let args = RootArgs::parse();

    smol::block_on(args.run());
    // async_std::task::block_on(args.run());
}
